// ipl_generatorGGL: A program for generation of structure reconstruction ILP in ibm clp format.
//
// The algorithm by K.Gorbunov, V.Lyubetsky, R.Gershgorin [http://lab6.iitp.ru/en/ilp_generatorGGL/]
//
// (C)2017, Roman Gershgorin, Institute for Information Transmission Problems RAS
// E-mail: gershr{at}mail.ru
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef STRUCTURE_H_
#define STRUCTURE_H_

#include "helpers.h"

#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>

struct DirectedGene
{
	int geneIndex;
	int paralogIndex;
	int sign;
};

struct Chromosome
{
	bool circular;
	std::vector<DirectedGene> chromosome;
};

inline DirectedGene StringToGene(const std::string& geneStr)
{
	DirectedGene gene;
	int start = 0;
	if (geneStr.front() == '-')
	{
		gene.sign = -1;
		start = 1;
	}
	else
		gene.sign = 1;

	int dotPos = geneStr.find('.');
	gene.geneIndex = FromString<int>(geneStr.substr(start, dotPos));

	if (dotPos != std::string::npos)
	{
		gene.paralogIndex = FromString<int>(geneStr.substr(dotPos + 1));
	}

	return gene;
}

struct Structure
{
	std::vector<Chromosome> chromosomes;
	std::map<int, std::vector<int> > availableParalogs;
	int id;
	std::string name;
};

using Structures = std::vector<Structure>;

inline std::stringstream ReplaceUselessSymbols(std::string& str)
{
	for (size_t index = 0; index != str.size(); ++index)
	{
		if (str[index] == ';' || str[index] == ':')
			str[index] = ' ';
	}

	return std::stringstream(str);
}


struct NameToIndexStorage
{
	std::map<std::string, int> storage;
	int minNewIndex;

	NameToIndexStorage()
		: minNewIndex(1)
	{}
};

inline void PrintNameIndex(std::ostream& out, const NameToIndexStorage& storage)
{
	for (const auto& entry : storage.storage)
	{
		out << entry.first << " " << entry.second << "\n";
	}
}

inline DirectedGene StringToGene1(const std::string& geneStr, NameToIndexStorage& storage)
{
	DirectedGene gene;
	int start = 0;

	int dotPos = geneStr.find('.');
	std::string geneName = geneStr.substr(start, dotPos);

	const auto nameIt = storage.storage.find(geneName);
	if (nameIt != storage.storage.end())
	{
		gene.geneIndex = nameIt->second;
	}
	else
	{
		gene.geneIndex = storage.minNewIndex;
		storage.storage[geneName] = storage.minNewIndex;
		++storage.minNewIndex;
	}

	if (dotPos != std::string::npos)
	{
		gene.paralogIndex = FromString<int>(geneStr.substr(dotPos + 1));
	}

	return gene;
}

inline void ReadChromosomeStructures(std::ifstream& in, Structures& structures, NameToIndexStorage& storage)
{
	int structuresNumber = 0;
	std::string tmp;
	while (tmp.empty())
	{
		std::getline(in, tmp);
	}
	ReplaceUselessSymbols(tmp) >> structuresNumber;

	int hash = 1;
	Structures structuresTmp;
	for (size_t i = 0; i != structuresNumber; ++i)
	{
		int chromosomeNumber = 0;
		std::map<std::string, int> count;
		Structure structure;
		std::getline(in, tmp);
		std::stringstream s = ReplaceUselessSymbols(tmp);
		s >> structure.name >> structure.id;
		s >> chromosomeNumber;
		for (size_t j = 0; j != chromosomeNumber; ++j)
		{
			Chromosome chromosome;

			std::string tmp1;
			s >> tmp1 >> tmp;
			char type = tmp1[0];
			chromosome.circular = (type == 'C') ? true : false;
			int geneNumber = FromString<int>(tmp1.substr(1));
			char signSymb;
			int index = 0;
			for (size_t k = 0; k != geneNumber; ++k)
			{
				std::string undirectedGeneName;
				while (tmp[index] == ' ')
				{
					++index;
				}
				signSymb = tmp[index++];
				int sign = (signSymb == '-') ? -1 : 1;
				while (index < tmp.size() && (tmp[index] != '-' && tmp[index] != '+' && tmp[index] != ' '))
				{
					undirectedGeneName += tmp[index++];
				}

				DirectedGene gene = StringToGene1(undirectedGeneName, storage);
				gene.sign = sign;
				chromosome.chromosome.push_back(gene);
				structure.availableParalogs[chromosome.chromosome.back().geneIndex].push_back(chromosome.chromosome.back().paralogIndex);
			}

			structure.chromosomes.push_back(chromosome);
		}
		structuresTmp.push_back(structure);
	}

	std::swap(structures, structuresTmp);
}

struct ParalogMax
{
	int geneIndex;
	int maxParalogIndex;
};

using ParalogsMaxParam = std::vector<ParalogMax>;

inline ParalogsMaxParam ReadParalogsMax(std::istream& input)
{
	int paramsNumber = 0;
	input >> paramsNumber;

	ParalogsMaxParam param;
	param.reserve(paramsNumber);
	for (size_t index = 0; index != paramsNumber; ++index)
	{
		ParalogMax p;
		input >> p.geneIndex >> p.maxParalogIndex;
		param.push_back(p);
	}

	return param;
}

#endif // STRUCTURE_H_
