// ipl_generatorGGL: A program for generation of structure reconstruction ILP in ibm clp format.
//
// The algorithm by K.Gorbunov, V.Lyubetsky, R.Gershgorin [http://lab6.iitp.ru/en/ilp_generatorGGL/]
//
// (C)2017, Roman Gershgorin, Institute for Information Transmission Problems RAS
// E-mail: gershr{at}mail.ru
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "task_generation.h"
#include "variables_storage.h"
#include "task_storage.h"

#include <map>
#include <boost/format.hpp>

namespace
{

const char* endSuffixes[] = { "_", "__" };
const char* GeneStartMark = "_";
const char* GeneEndMark = "__";

//const char* endSuffixes[] = { "\'", "\'\'" };
//const char* GeneStartMark = "\'";
//const char* GeneEndMark = "\'\'";

struct GeneEnd
{
	Gene gene;
	std::string end;

	GeneEnd()
	{}

	GeneEnd(const Gene& _gene, const std::string& _end)
		: gene(_gene)
		, end(_end)
	{}

	bool operator==(const GeneEnd& other) const
	{
		return gene == other.gene && end == other.end;
	}

	bool EqualGene(const GeneEnd& other) const
	{
		return gene.gene == other.gene.gene;
	}

	std::string Str() const
	{
		return gene.Str() + end;
	}
};

std::vector<GeneEnd> GetPossibleEnds(const ParalogsMaxParam& paralogsMax)
{
	std::vector<GeneEnd> ends;

	const std::vector<std::string> endMarks = { GeneStartMark, GeneEndMark };

	for (const auto& gene : paralogsMax)
	{
		for (int par = 1; par != gene.maxParalogIndex + 1; ++par)
		{
			for (const auto& mark : endMarks)
			{
				ends.push_back(GeneEnd(Gene(gene.geneIndex, par), mark));
			}
		}
	}

	return ends;
}

bool IsLeaf(const Node& node)
{
	return node.leftSonRef == -1 && node.rightSonRef == -1;
}

bool ArePotentiallyEqualLeafEdges(const GeneEnd& end1, const GeneEnd& end2)
{
	return (end1.end == end2.end) && end1.EqualGene(end2);
}

bool ArePotentiallyEqualInnerEdges(const GeneEnd& end1, const GeneEnd& end2)
{
	return end1 == end2;
}

Structure GetStructureByNodeId(const Structures& structures, int nodeId)
{
	for (int i = 0; i != structures.size(); ++i)
	{
		if (structures[i].id == nodeId)
			return structures[i];
	}

	throw std::runtime_error("Failed to get structure by nodeId");
}

} // namespace

void PrintZVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, VariablesStorage& vars)
{
	std::map<int, int> params;
	for (const auto& p : paralogsMax)
	{
		params[p.geneIndex] = p.maxParalogIndex;
	}

	for (const auto& structs : structures)
	{
		int structId = structs.id;

		for (const auto& chromosome : structs.chromosomes)
		{
			std::map<std::string,std::string> zVarBoundI;
			for (const auto& gene : chromosome.chromosome)
			{
				std::string zVarBoundJ;
				for (size_t index = 1; index != params[gene.geneIndex] + 1; ++index)
				{
					const auto z = vars.AddZVariable(structId, gene.geneIndex, gene.paralogIndex, index);
					zVarBoundJ += z + (index == params[gene.geneIndex] ? " " : " + ");
					const auto ind = ToString(gene.geneIndex) + '.' + ToString(index);
					zVarBoundI[ind] += (zVarBoundI[ind].empty() ? "" : " + ") + z;				
				}
				zVarBoundJ += "= 1";
				print(zVarBoundJ);
				AddStringToStorage(storage, "Subject To", zVarBoundJ);
			}
			for (auto zVarBoundIS : zVarBoundI)
			{
				if (!zVarBoundIS.second.empty())
				{
					print(zVarBoundIS.second + " <= 1");
					AddStringToStorage(storage, "Subject To", zVarBoundIS.second + " <= 1");
				}
			}
		}
	}
	
}

void PrintXVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::string z, x;

	const auto ends = GetPossibleEnds(paralogsMax);

	for (const auto& node : tree)
	{
		if (IsLeaf(node))
		{
			continue;
		}

		for (int i = 0; i != ends.size(); ++i)
		{
			std::string xVarBound = "";
			const auto p1 = ends[i];

			for (int j = 0; j != ends.size(); ++j)
			{
				const auto p2 = ends[j];
				if (p1 == p2)
				{
					continue;
				}

				x = vars.AddXVariable(node.nodeId, p1.Str(), p2.Str());
				xVarBound += (xVarBound.empty() ? "" : " + ") + x;
			}
			const auto y = vars.AddYVariable(node.nodeId, ends[i].gene);
			print(xVarBound + " + " + y + " = 1");
			AddStringToStorage(storage, "Subject To", xVarBound + " + " + y + " = 1");
		}
	}
}

void PrintYVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::string x, y, z;
	for (const auto& node : tree)
	{
		if (IsLeaf(node))
		{
			continue;
		}

		for (int i = 0; i != paralogsMax.size(); ++i)
		{
			for (int j = 1; j != paralogsMax[i].maxParalogIndex + 1; ++j)
			{
				const auto y = vars.AddYVariable(node.nodeId, Gene(paralogsMax[i].geneIndex, j));
				std::string yVarBound = y;

				const auto v1 = tree[node.leftSonRef];
				const auto v1Id = ToString(v1.nodeId);
				int rhsValue = 0;
				if (!IsLeaf(v1))
				{
					const auto y1 = vars.AddYVariable(v1.nodeId, Gene(paralogsMax[i].geneIndex, j));
					yVarBound += " - " + y1;
				}
				else
				{
					auto v1Str = GetStructureByNodeId(structures, v1.nodeId);
					rhsValue += 1;
					for (int k = 0; k != v1Str.availableParalogs[paralogsMax[i].geneIndex].size(); ++k)
					{
						const auto zk = vars.AddZVariable(v1.nodeId, paralogsMax[i].geneIndex, v1Str.availableParalogs[paralogsMax[i].geneIndex][k], j);
						yVarBound += " + " + zk;
					}
				}

				const auto v2 = tree[node.rightSonRef];
				const auto v2Id = ToString(v2.nodeId);
				if (!IsLeaf(v2))
				{
					const auto y2 = vars.AddYVariable(v2.nodeId, Gene(paralogsMax[i].geneIndex, j));
					yVarBound += " - "  + y2;
				}
				else
				{
					auto v2Str = GetStructureByNodeId(structures, v2.nodeId);
					rhsValue += 1;
					for (int k = 0; k != v2Str.availableParalogs[paralogsMax[i].geneIndex].size(); ++k)
					{
						const auto zk = vars.AddZVariable(v2.nodeId, paralogsMax[i].geneIndex, v2Str.availableParalogs[paralogsMax[i].geneIndex][k], j);
						yVarBound += " + " + zk;
					}

				}

				yVarBound += " <= " + ToString(rhsValue);

				print(yVarBound);
				AddStringToStorage(storage, "Subject To", yVarBound);
			}
		}
	}
}

std::vector<std::pair<GeneEnd, GeneEnd> > GetIncidentEnds(const Structure& structure)
{
	std::vector<std::pair<GeneEnd, GeneEnd> > incidentEnds;

	for (const auto& c : structure.chromosomes)
	{
		for (int i = 0; i != c.chromosome.size() - 1; ++i)
		{
			GeneEnd first;
			first.gene.gene = c.chromosome[i].geneIndex;
			first.gene.par = c.chromosome[i].paralogIndex;
			first.end = c.chromosome[i].sign == 1 ? GeneEndMark : GeneStartMark;

			GeneEnd second;
			second.gene.gene = c.chromosome[i + 1].geneIndex;
			second.gene.par = c.chromosome[i + 1].paralogIndex;
			second.end = c.chromosome[i + 1].sign == 1 ? GeneStartMark : GeneEndMark;

			incidentEnds.push_back(std::make_pair(first, second));
		}

		if (c.circular)
		{
			GeneEnd first;
			first.gene.gene = c.chromosome.back().geneIndex;
			first.gene.par = c.chromosome.back().paralogIndex;
			first.end = c.chromosome.back().sign == 1 ? GeneEndMark : GeneStartMark;

			GeneEnd second;
			second.gene.gene = c.chromosome.front().geneIndex;
			second.gene.par = c.chromosome.front().paralogIndex;
			second.end = c.chromosome.front().sign == 1 ? GeneStartMark : GeneEndMark;

			incidentEnds.push_back(std::make_pair(first, second));
		}
	}

	return incidentEnds;
}

void PrintTForIncidentEnds(Storage& storage, PrinterType& print, const Structure& structure, const std::map<int, int>& params, int aId, int bId, VariablesStorage& vars)
{
	const auto incidentEnds = GetIncidentEnds(structure);

	for (const auto& pair : incidentEnds)
	{
		auto p1 = pair.first;
		auto p2 = pair.second;

		for (int l1 = 1; l1 != params.at(p1.gene.gene) + 1; ++l1)
		{
			for (int l2 = 1; l2 != params.at(p2.gene.gene) + 1; ++l2)
			{
				std::string y_coef1 = ToString(p1.gene.gene) + "." + ToString(l1);
				std::string y_coef2 = ToString(p2.gene.gene) + "." + ToString(l2);

				if (y_coef1 == y_coef2)
					continue;

				const auto t = vars.AddTVariable(bId, bId, p1.Str(), p2.Str());
				const auto y1 = vars.AddYVariable(aId, Gene(p1.gene.gene, l1));
				const auto y2 = vars.AddYVariable(aId, Gene(p2.gene.gene, l2));
				const auto z1 = vars.AddZVariable(bId, p1.gene.gene, p1.gene.par, l1);
				const auto z2 = vars.AddZVariable(bId, p2.gene.gene, p2.gene.par, l2);

				const std::string tVarInc1 = y1 + " - " + y2 + " + " + z1 + " + " + z2 + " - " + t + " <= 2";
				const std::string tVarInc2 = y2 + " - " + y1 + " + " + z1 + " + " + z2 + " - " + t + " <= 2";

				print(tVarInc1);
				AddStringToStorage(storage, "Subject To", tVarInc1);
				print(tVarInc2);
				AddStringToStorage(storage, "Subject To", tVarInc2);
			}
		}
	}
}


void PrintTVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::map<int, int> params;
	for (const auto& p : paralogsMax)
	{
		params[p.geneIndex] = p.maxParalogIndex;
	}

	for (const auto& bNode : tree)
	{
		if (bNode.parentRef == -1)
		{
			continue;
		}

		const auto& aNode = tree[bNode.parentRef];
		
		if (!IsLeaf(bNode))
		{
			const auto ends = GetPossibleEnds(paralogsMax);

			for (int p = 0; p != ends.size(); ++p)
			{
				for (int q = 0; q != ends.size(); ++q)
				{
					auto p1 = ends[p];
					auto p2 = ends[q];

					if (p1 == p2)
						continue;

					const auto ta = vars.AddTVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto x1 = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());
					const auto y_b11 = vars.AddYVariable(bNode.nodeId, p1.gene);
					const auto y_b22 = vars.AddYVariable(bNode.nodeId, p2.gene);
					std::string tVarBound1 = ta + " - " + x1;
					std::string tVarBound2 = tVarBound1;

					tVarBound1 += " - " + y_b11 + " + " + y_b22 + " >= -1";
					tVarBound2 += " - " + y_b22 + " + " + y_b11 + " >= -1";

					print(tVarBound1);
					AddStringToStorage(storage, "Subject To", tVarBound1);
					print(tVarBound2);
					AddStringToStorage(storage, "Subject To", tVarBound2);

					const auto tb = vars.AddTVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto x2 = vars.AddXVariable(bNode.nodeId, p1.Str(), p2.Str());
					const auto y_a11 = vars.AddYVariable(aNode.nodeId, p1.gene);
					const auto y_a22 = vars.AddYVariable(aNode.nodeId, p2.gene);

					std::string tVarBound3 = tb + " - " + x2;
					std::string tVarBound4 = tVarBound3;

					tVarBound3 += " - " + y_a11 + " + " + y_a22 + " >= -1";
					tVarBound4 += " - " + y_a22 + " + " + y_a11 + " >= -1";

					print(tVarBound3);
					AddStringToStorage(storage, "Subject To", tVarBound3);
					print(tVarBound4);
					AddStringToStorage(storage, "Subject To", tVarBound4);
				}
			}
			continue;
		}

		auto bStruct = GetStructureByNodeId(structures, bNode.nodeId);
		const auto ends = GetPossibleEnds(paralogsMax);

		PrintTForIncidentEnds(storage, print, bStruct, params, aNode.nodeId, bNode.nodeId, vars);

		for (int p = 0; p != ends.size(); ++p)
		{
			for (int q = 0; q != ends.size(); ++q)
			{
				auto p1 = ends[p];
				auto p2 = ends[q];

				if (p1 == p2)
					continue;

				const auto t = vars.AddTVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
				const auto x = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());

				std::string tVarBound1 = x + " - " + t;
				std::string tVarBound2 = tVarBound1;

				for (int i = 0; i != bStruct.availableParalogs[p1.gene.gene].size(); ++i)
				{
					const auto z = vars.AddZVariable(bNode.nodeId, p1.gene.gene, bStruct.availableParalogs[p1.gene.gene][i], p1.gene.par);
					tVarBound1 += " + " + z;
					tVarBound2 += " - " + z;
				}
				for (int i = 0; i != bStruct.availableParalogs[p2.gene.gene].size(); ++i)
				{
					const auto z = vars.AddZVariable(bNode.nodeId, p2.gene.gene, bStruct.availableParalogs[p2.gene.gene][i], p2.gene.par);
					tVarBound1 += " - " + z;
					tVarBound2 += " + " + z;
				}

				print(tVarBound1 + " <= 1");
				AddStringToStorage(storage, "Subject To", tVarBound1 + " <= 1");
				print(tVarBound2 + " <= 1");
				AddStringToStorage(storage, "Subject To", tVarBound2 + " <= 1");
			}
		}
	}
}

void PrintRVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::map<int, int> params;
	for (const auto& p : paralogsMax)
	{
		params[p.geneIndex] = p.maxParalogIndex;
	}

	for (const auto& bNode : tree)
	{
		if (bNode.parentRef == -1)
		{
			continue;
		}

		const auto aNode = tree[bNode.parentRef];
		const auto ends = GetPossibleEnds(paralogsMax);

		if (!IsLeaf(bNode))
		{
			for (int p1i = 0; p1i < ends.size(); ++p1i)
			{
				for (int p2i = 0; p2i < ends.size(); ++p2i)
				{
					for (int p3i = 0; p3i < ends.size(); ++p3i)
					{
						if (p1i == p2i || p1i == p3i)
							continue;

						auto p1 = ends[p1i];
						auto p2 = ends[p2i];
						auto p3 = ends[p3i];

						auto end1 = p1.Str();
						auto end2 = p2.Str();
						auto end3 = p3.Str();

						const auto ra = vars.AddRVariable(aNode.nodeId, bNode.nodeId, end1, end2);
						const auto xa = vars.AddXVariable(aNode.nodeId, end1, end2);
						const auto xb = vars.AddXVariable(bNode.nodeId, end1, end3);
						const auto yb = vars.AddYVariable(bNode.nodeId, p2.gene);
						const auto ya = vars.AddYVariable(aNode.nodeId, p3.gene);
						const auto rb = vars.AddRVariable(bNode.nodeId, bNode.nodeId, end1, end3);

						std::string rVarBound1 = ra + " + " + xa + " + " + xb + " - " + yb + " - " + ya + " + " + rb + " <= 3";
						std::string rVarBound2 = xa + " + " + xb + " - " + yb + " - " + ya + " - " + rb + " - " + ra + " <= 1";

						print(rVarBound1);
						AddStringToStorage(storage, "Subject To", rVarBound1);
						print(rVarBound2);
						AddStringToStorage(storage, "Subject To", rVarBound2);
					}
				}
			}

			continue;
		}

		auto bStruct = GetStructureByNodeId(structures, bNode.nodeId);

		const auto incidentEnds = GetIncidentEnds(bStruct);

		auto incidentEndsAll = incidentEnds;
		for (auto p : incidentEnds)
		{
			std::swap(p.first, p.second);
			incidentEndsAll.push_back(p);
		}

		for (int p = 0; p != incidentEndsAll.size(); ++p)
		{
			for (int p1i = 0; p1i != ends.size(); ++p1i)
			{
				for (int p2i = 0; p2i != ends.size(); ++p2i)
				{
					auto p1 = ends[p1i];
					auto p2 = ends[p2i];
					auto p1p = incidentEndsAll[p].first;
					auto p3 = incidentEndsAll[p].second;

					if (!ArePotentiallyEqualLeafEdges(p1, p1p) || (p1 == p2) || (p1p == p3))
					{			
						continue;
					}

					const auto ra = vars.AddRVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto rb = vars.AddRVariable(bNode.nodeId, bNode.nodeId, p1p.Str(), p3.Str());
					const auto x = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());
					const auto z = vars.AddZVariable(bNode.nodeId, p1.gene.gene, p1p.gene.par, p1.gene.par);

					std::string rVarBound1 = ra + " + " + x + " + " + z;

					for (int i = 0; i != bStruct.availableParalogs[p2.gene.gene].size(); ++i)
					{
						const auto zi = vars.AddZVariable(bNode.nodeId, p2.gene.gene, bStruct.availableParalogs[p2.gene.gene][i], p2.gene.par);
						rVarBound1 += " + " + zi;
					}

					for (int i = 1; i != params[p3.gene.gene] + 1; ++i)
					{
						const auto yt = vars.AddYVariable(aNode.nodeId, Gene(p3.gene.gene, i));
						const auto zt = vars.AddZVariable(bNode.nodeId, p3.gene.gene, p3.gene.par, i);

						print(rVarBound1 + " - " + yt + " + " + zt + " + " + rb + " <= 5");
						AddStringToStorage(storage, "Subject To", rVarBound1 + " - " + yt + " + " + zt + " + " + rb + " <= 5");
					}

					std::string rVarBound2 = x + " + " + z;

					for (int i = 0; i != bStruct.availableParalogs[p2.gene.gene].size(); ++i)
					{
						const auto zi = vars.AddZVariable(bNode.nodeId, p2.gene.gene, bStruct.availableParalogs[p2.gene.gene][i], p2.gene.par);
						rVarBound2 += " + " + zi;
					}

					for (int i = 1; i != params[p3.gene.gene] + 1; ++i)
					{
						const auto yt = vars.AddYVariable(aNode.nodeId, Gene(p3.gene.gene, i));
						const auto zt = vars.AddZVariable(bNode.nodeId, p3.gene.gene, p3.gene.par, i);

						print(rVarBound2 + " - " + yt + " + " + zt + " - " + rb + " - " + ra + " <= 3");
						AddStringToStorage(storage, "Subject To", rVarBound2 + " - " + yt + " + " + zt + " - " + rb + " - " + ra + " <= 3");
					}
				}
			}
		}
	}
}

void PrintUVariables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::map<int, int> params;
	for (const auto& p : paralogsMax)
	{
		params[p.geneIndex] = p.maxParalogIndex;
	}

	const std::vector<GeneEnd> ends = GetPossibleEnds(paralogsMax);

	for (const auto& bNode : tree)
	{
		int mValue = 1;
		if (bNode.parentRef == -1)
		{
			continue;
		}

		const auto aNode = tree[bNode.parentRef];

		if (!IsLeaf(bNode))
		{
			for (int p1i = 0; p1i != ends.size() - 1; ++p1i)
			{
				for (int p2i = p1i + 1; p2i != ends.size(); ++p2i)
				{
					auto p1 = ends[p1i];
					auto p2 = ends[p2i];

					const auto ua = vars.AddUVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto ma = vars.AddMVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
					const auto xa = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());

					std::string uVarBound1 = ma + " " + xa + " - " + ua + " >= 0" ;
					print(uVarBound1);
					AddStringToStorage(storage, "Subject To", uVarBound1);

					const auto y1 = vars.AddYVariable(bNode.nodeId, p1.gene);
					std::string uVarBound2 = ua + " + " + ma + " " + y1 + " <= " + ma;
					print(uVarBound2);
					AddStringToStorage(storage, "Subject To", uVarBound2);

					const auto y2 = vars.AddYVariable(bNode.nodeId, p2.gene);
					std::string uVarBound3 = ua + " + " + ma + " " + y2 + " <= " + ma;
					print(uVarBound3);
					AddStringToStorage(storage, "Subject To", uVarBound3);

					const auto ub = vars.AddUVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto mb = vars.AddMVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
					const auto xb = vars.AddXVariable(bNode.nodeId, p1.Str(), p2.Str());

					std::string uVarBound4 = mb + " " + xb + " - " + ub + " >= 0";
					print(uVarBound4);
					AddStringToStorage(storage, "Subject To", uVarBound4);

					const auto y3 = vars.AddYVariable(aNode.nodeId, p1.gene);
					std::string uVarBound5 = ub + " + " + mb + " " + y3 + " <= " + mb;
					print(uVarBound5);
					AddStringToStorage(storage, "Subject To", uVarBound5);

					const auto y4 = vars.AddYVariable(aNode.nodeId, p2.gene);
					std::string uVarBound6 = ub + " + " + mb + " " + y4 + " <= " + mb;
					print(uVarBound6);
					AddStringToStorage(storage, "Subject To", uVarBound6);
				}
			}

			continue;
		}

		auto bStruct = GetStructureByNodeId(structures, bNode.nodeId);

		for (int p1i = 0; p1i != ends.size() - 1; ++p1i)
		{
			for (int p2i = p1i + 1; p2i != ends.size(); ++p2i)
			{
				auto p1 = ends[p1i];
				auto p2 = ends[p2i];

				const auto ua = vars.AddUVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
				const auto ma = vars.AddMVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
				const auto xa = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());

				std::string uVarBound1 = ma + " " + xa + " - " + ua + " >= 0";
				print(uVarBound1);
				AddStringToStorage(storage, "Subject To", uVarBound1);

				std::string uVarBound2 = "";
				for (int i = 0; i != bStruct.availableParalogs[p1.gene.gene].size(); ++i)
				{
					const auto zi = vars.AddZVariable(bNode.nodeId, p1.gene.gene, bStruct.availableParalogs[p1.gene.gene][i], p1.gene.par);
					uVarBound2 += ma + " " + zi;
					if (i < bStruct.availableParalogs[p1.gene.gene].size() - 1)
					{
						uVarBound2 += " + ";
					}
				}
				print(uVarBound2 + " - " + ua + " >= 0");
				AddStringToStorage(storage, "Subject To", uVarBound2 + " - " + ua + " >= 0");

				std::string uVarBound3 = "";
				for (int i = 0; i != bStruct.availableParalogs[p2.gene.gene].size(); ++i)
				{
					const auto zi = vars.AddZVariable(bNode.nodeId, p2.gene.gene, bStruct.availableParalogs[p2.gene.gene][i], p2.gene.par);
					uVarBound3 += ma + " " + zi;
					if (i < bStruct.availableParalogs[p2.gene.gene].size() - 1)
					{
						uVarBound3 += " + ";
					}
				}
				print(uVarBound3 + " - " + ua + " >= 0");
				AddStringToStorage(storage, "Subject To", uVarBound3 + " - " + ua + " >= 0");
			}
		}

		const auto incidentEnds = GetIncidentEnds(bStruct);
		for (const auto& p : incidentEnds)
		{
			auto p1 = p.first;
			auto p2 = p.second;

			if (p1 == p2)
				continue;

			const auto ub = vars.AddUVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
			const auto mb = vars.AddMVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
			const auto xb = vars.AddXVariable(bNode.nodeId, p1.Str(), p2.Str());

			std::string uVarBound4 = mb + " " + xb + " - " + ub + " >= 0";
			print(uVarBound4);
			AddStringToStorage(storage, "Subject To", uVarBound4);
			print(xb + " = 1");
			AddStringToStorage(storage, "Subject To", xb + " = 1");



			std::string uVarBound5 = ub;
			for (int i = 1; i != params[p1.gene.gene] + 1; ++i)
			{
				const auto zi = vars.AddZVariable(bNode.nodeId, p1.gene.gene, p1.gene.par, i);
				const auto yi = vars.AddYVariable(aNode.nodeId, Gene(p1.gene.gene, i));

				print(uVarBound5 + " + " + mb + " " + zi + " + " + mb + " " + yi + " <= " + ToString(2 * ToInt(mb)));
				AddStringToStorage(storage, "Subject To", uVarBound5 + " + " + mb + " " + zi + " + " + mb + " " + yi + " <= " + ToString(2 * ToInt(mb)));
			}

			std::string uVarBound6 = ub;

			for (int i = 1; i != params[p2.gene.gene] + 1; ++i)
			{
				const auto zi = vars.AddZVariable(bNode.nodeId, p2.gene.gene, p2.gene.par, i);
				const auto yi = vars.AddYVariable(aNode.nodeId, Gene(p2.gene.gene, i));

				print(uVarBound6 + " + " + mb + " " + zi + " + " + mb + " " + yi + " <= " + ToString(2 * ToInt(mb)));
				AddStringToStorage(storage, "Subject To", uVarBound6 + " + " + mb + " " + zi + " + " + mb + " " + yi + " <= " + ToString(2 * ToInt(mb)));
			}
		}
	}
}

void PrintU1Variables(Storage& storage, PrinterType& print, const Structures& structures, const ParalogsMaxParam& paralogsMax, const Tree& tree, VariablesStorage& vars)
{
	std::map<int, int> params;
	for (const auto& p : paralogsMax)
	{
		params[p.geneIndex] = p.maxParalogIndex;
	}

	for (const auto& bNode : tree)
	{
		if (bNode.parentRef == -1)
		{
			continue;
		}

		int mValue = 1;

		const auto aNode = tree[bNode.parentRef];
		const auto ends = GetPossibleEnds(paralogsMax);

		if (!IsLeaf(bNode))
		{
			for (int p1i = 0; p1i != ends.size(); ++p1i)
			{
				for (int p2i = 0; p2i != ends.size(); ++p2i)
				{
					for (int p3i = 0; p3i != ends.size(); ++p3i)
					{
						auto p1 = ends[p1i];
						auto p2 = ends[p2i];
						auto p3 = ends[p3i];

						if (p1 == p2 || p1 == p3)
						{
							continue;
						}

						const auto ua = vars.AddUVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
						const auto ub = vars.AddUVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p3.Str());
						const auto ma = vars.AddMVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
						const auto mb = vars.AddMVariable(bNode.nodeId, bNode.nodeId, p1.Str(), p3.Str(), mValue);
						const auto xa = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());
						const auto xb = vars.AddXVariable(bNode.nodeId, p1.Str(), p3.Str());

						std::string uVarBound1 = ua + " + " + ma + " " + xb + " - " + ub + " <= " + ma;
						std::string uVarBound2 = ub + " + " + mb + " " + xa + " - " + ua + " <= " + mb;

						print(uVarBound1);
						AddStringToStorage(storage, "Subject To", uVarBound1);
						print(uVarBound2);
						AddStringToStorage(storage, "Subject To", uVarBound2);
					}
				}
			}
			continue;
		}

		auto bStruct = GetStructureByNodeId(structures, bNode.nodeId);

		const auto incidentEnds = GetIncidentEnds(bStruct);

		auto incidentEndsAll = incidentEnds;
		for (auto p : incidentEnds)
		{
			std::swap(p.first, p.second);
			incidentEndsAll.push_back(p);
		}

		for (int p = 0; p != incidentEndsAll.size(); ++p)
		{
			for (int p1i = 0; p1i != ends.size(); ++p1i)
			{
				for (int p2i = 0; p2i != ends.size(); ++p2i)
				{
					auto p1 = ends[p1i];
					auto p2 = ends[p2i];
					auto p1p = incidentEndsAll[p].first;
					auto p3 = incidentEndsAll[p].second;

					if (!ArePotentiallyEqualLeafEdges(p1, p1p) || p1 == p2 || p1p == p3)
					{
						continue;
					}

					const auto ua = vars.AddUVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str());
					const auto ub = vars.AddUVariable(bNode.nodeId, bNode.nodeId, p1p.Str(), p3.Str());
					const auto ma = vars.AddMVariable(aNode.nodeId, bNode.nodeId, p1.Str(), p2.Str(), mValue);
					const auto mb = vars.AddMVariable(bNode.nodeId, bNode.nodeId, p1p.Str(), p3.Str(), mValue);
					const auto z = vars.AddZVariable(bNode.nodeId, p1.gene.gene, p1p.gene.par, p1.gene.par);
					const auto x = vars.AddXVariable(aNode.nodeId, p1.Str(), p2.Str());

					std::string uVarBound1 = ua + " - " + ub + " + " + ma + " " + z + " <= " + ma;
					std::string uVarBound2 = ub + " - " + ua + " + " + mb + " " + x + " + " + mb + " " + z + " <= " + ToString(2 * ToInt(mb));
							
					print(uVarBound1);
					AddStringToStorage(storage, "Subject To", uVarBound1);
					print(uVarBound2);
					AddStringToStorage(storage, "Subject To", uVarBound2);
				}
			}
		}
	}
}

void PrintPVariables(Storage& storage, PrinterType& print, VariablesStorage& vars)
{
	const auto genVars = vars.GetMVars();

	for (const auto& var : genVars)
	{
		const auto p = vars.AddPVariable(var.first);
		auto u = var.first;
		u[0] = 'u';
		print(u + " - " + ToString(genVars.at(var.first)) + " " + p + " >= 0");
		AddStringToStorage(storage, "Subject To", u + " - " + ToString(genVars.at(var.first)) + " " + p + " >= 0");
	}
}
