// ipl_generatorGGL: A program for generation of structure reconstruction ILP in ibm clp format.
//
// The algorithm by K.Gorbunov, V.Lyubetsky, R.Gershgorin [http://lab6.iitp.ru/en/ilp_generatorGGL/]
//
// (C)2017, Roman Gershgorin, Institute for Information Transmission Problems RAS
// E-mail: gershr{at}mail.ru
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "task_storage.h"
#include "helpers.h"

#include <boost/range/iterator_range.hpp>
#include <boost/algorithm/string.hpp>

namespace
{

struct Monom
{
	std::string var;
	int coef;
};

bool IsSign(char s)
{
	return (s == '-') || (s == '+');
}

bool IsSign(const std::string& s, int& sign)
{
	if (s.empty() || !IsSign(s.front()))
	{
		return false;
	}

	sign = (s.front() == '-' ? -1 : 1);
	return true;
}

bool IsVar(const std::string& token)
{
	return token[0] >= 'a' && token[0] <= 'z';
}

} // namespace

std::string BuildExpression(const std::map<std::string, int>& polynom)
{
	std::string expression;

	for (auto it = polynom.begin(); it != polynom.end(); ++it)
	{
		expression += ((it->second > 0) ? "+" : "-");
		if (abs(it->second) != 1)
			expression += " " + ToString(abs(it->second));
		expression += " " + it->first + " ";
	}

	boost::trim(expression);
	if (expression.front() == '+')
	{
		expression = expression.substr(2, expression.size() - 2);
	}

	return expression;
}

struct Comparator
{
	bool operator()(const std::pair<std::string, int>& p)
	{
		return p.second == 0;
	}
};

std::map<std::string, int> ParseExpression(const std::string& expression)
{
	std::string tmp = expression;

	if (!IsSign(tmp.front()))
	{
		tmp = "+ " + tmp;
	}

	std::map<std::string, int> polynom;

	using SplitIterator = boost::split_iterator<std::string::iterator>;

	Monom monom;

	for (
		SplitIterator it = boost::make_split_iterator(tmp, boost::first_finder(" ", boost::is_equal()));
		it != SplitIterator();
	++it)
	{
		int sign = 1;
		const std::string token = boost::copy_range<std::string>(*it);
		if (IsSign(token, sign))
		{
			monom.coef = sign;
		}
		else if (IsVar(token))
		{
			monom.var = token;
			auto it = polynom.find(monom.var);

			if (it == polynom.end())
			{
				polynom[monom.var] = monom.coef;
			}
			else
			{
				it->second += monom.coef;
			}
		}
		else
		{
			monom.coef *= FromString<int>(token);
		}
	}

	for (auto it = polynom.begin(); it != polynom.end(); )
		if (it->second == 0)
			it = polynom.erase(it);
		else
			++it;

	return polynom;
}

Inequality ParseInequality(const std::string& expression)
{
	std::string::size_type pos = expression.find('=');
	if (pos == std::string::npos || pos == 0)
	{
		throw std::runtime_error("Provided expression is invalid inequality");
	}
	std::string::size_type lhsEnd = (expression[pos - 1] != ' ' ? pos - 2 : pos - 1);
	std::string::size_type rhsBegin = pos + 2;

	Inequality inequality;
	inequality.lhs = ParseExpression(expression.substr(0, lhsEnd));
	inequality.senseOfInequality = expression.substr(lhsEnd + 1, rhsBegin - lhsEnd - 2);
	inequality.rhs = FromString<int>(expression.substr(rhsBegin, expression.size() - rhsBegin + 1));
	return inequality;
}

std::string BuildInequality(const Inequality& inequality)
{
	std::string expression = BuildExpression(inequality.lhs);
	expression += " " + inequality.senseOfInequality + " " + ToString(inequality.rhs);

	return expression;
}

void AddStringToStorage(Storage& storage, const std::string& section, const std::string& str)
{
	//std::cout << "Adding expression:\n" << str << "\n";

	auto inequality = BuildInequality(ParseInequality(str));
	storage[section].push_back(inequality);

	//std::cout << "Added:\n" << inequality << "\n";
}

void PrintTaskFromStorage(std::ostream& out, const Storage& storage)
{
	const std::vector<std::string> order = { "Minimize", "Subject To", "Bounds", "General", "Binary" };

	for (const auto& section : order)
	{
		for (auto& sectionIt : storage)
		{
			if (sectionIt.first == section)
			{
				out << sectionIt.first << "\n";
				for (auto& exprIt : sectionIt.second)
				{
					out << "\t" << exprIt << "\n";
				}
				break;
			}
		}
	}

}
