// ipl_generatorGGL: A program for generation of structure reconstruction ILP in ibm clp format.
//
// The algorithm by K.Gorbunov, V.Lyubetsky, R.Gershgorin [http://lab6.iitp.ru/en/ilp_generatorGGL/]
//
// (C)2017, Roman Gershgorin, Institute for Information Transmission Problems RAS
// E-mail: gershr{at}mail.ru
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef _VARIABLES_STORAGE_H_
#define _VARIABLES_STORAGE_H_

#include "helpers.h"
#include "task_storage.h"

#include <vector>
#include <set>
#include <string>
#include <map>
#include <ostream>

inline int ToInt(const std::string& val)
{
	return FromString<int>(val);
}

struct Gene
{
	int gene;
	int par;

	Gene()
	{}

	Gene(int gene, int par)
		: gene(gene)
		, par(par)
	{}

	Gene& operator=(const Gene& other)
	{
		if (this == &other)
			return *this;

		gene = other.gene;
		par = other.par;
		return *this;
	}

	bool operator==(const Gene& other) const
	{
		return gene == other.gene && par == other.par;
	}

	std::string Str() const
	{
		return ToString(gene) + "." + ToString(par);
	}
};

class VariablesStorage
{
public:
	std::string AddZVariable(int v, int k, int i, int j)
	{
		const std::string var = "z_" + ToString(v) + "." + ToString(k) + "." + ToString(i) + "." + ToString(j);
		m_binVars.insert(var);
		return var;
	}

	std::string AddXVariable(int v, std::string firstEnd, std::string secondEnd)
	{
		if (firstEnd > secondEnd)
			std::swap(firstEnd, secondEnd);

		const std::string var = "x_" + ToString(v) + "&" + firstEnd + secondEnd;
		m_binVars.insert(var);
		return var;
	}

	std::string AddTVariable(int v1, int v2, std::string firstEnd, std::string secondEnd)
	{
		if (firstEnd > secondEnd)
			std::swap(firstEnd, secondEnd);

		const std::string var = "t_" + ToString(v1) + "_" + ToString(v2) + "&" + firstEnd + secondEnd;
		m_binVars.insert(var);
		return var;
	}

	std::string AddYVariable(int v, const Gene& end)
	{
		const std::string var = "y_" +ToString(v) + "&" + ToString(end.gene) + "." + ToString(end.par);
		m_binVars.insert(var);
		return var;
	}

	std::string AddRVariable(int v1, int v2, std::string firstEnd, std::string secondEnd)
	{
		if (firstEnd > secondEnd)
			std::swap(firstEnd, secondEnd);

		const std::string var = "r_" + ToString(v1) + "_" + ToString(v2) + "&" + firstEnd + secondEnd;
		m_binVars.insert(var);
		return var;
	}

	std::string AddUVariable(int v1, int v2, std::string firstEnd, std::string secondEnd)
	{
		if (firstEnd > secondEnd)
			std::swap(firstEnd, secondEnd);

		const std::string var = "u_" + ToString(v1) + "_" + ToString(v2) + "&" + firstEnd + secondEnd;
		m_genVars.insert(var);
		return var;
	}

	std::string AddMVariable(int v1, int v2, std::string firstEnd, std::string secondEnd, int& value)
	{
		if (firstEnd > secondEnd)
			std::swap(firstEnd, secondEnd);

		const std::string mNode = "m_" + ToString(v2);
		const std::string var = "m_" + ToString(v1) + "_" + ToString(v2) + "&" + firstEnd + secondEnd;
		
		if (m_byNodeVals.find(mNode) == m_byNodeVals.end())
		{
			m_byNodeVals[mNode] = 1;
		}

		if (m_genVars.find(var) == m_genVars.end())
		{
			m_genVars.insert(var);
			m_mVars[var] = m_byNodeVals[mNode]++;
		}

		return ToString(m_mVars[var]);
	}

	std::string AddPVariable(std::string var)
	{
		var[0] = 'p';
		m_binVars.insert(var);
		return var;
	}

	void PrintData(std::ostream& out, Storage& storage) const
	{
		out << "Minimize\n";

		PrintMinimizationFunctional(out);

		out << "\n";

		PrintTaskFromStorage(out, storage);

		out << "\n";

		out << "Bounds\n";

		for (const auto& var : m_genVars)
		{
			if (var.front() == 'u')
			{
				auto m = var;
				m[0] = 'm';
				out << "\t0 <= " << var << " <= " + ToString(m_mVars.at(m)) + "\n";
			}
		}

		out << "\n";

		out << "General\n";

		for (const auto& var : m_genVars)
		{
			if (var.front() != 'm')
			{
				out << "\t" << var << "\n";
			}
		}
		out << "\n";
		out << "Binary\n";

		for (const auto& var : m_binVars)
		{
			out << "\t" << var << "\n";
		}

		out << "\n";
		out << "End";
	}

	std::map<std::string, int> GetMVars() const
	{
		return m_mVars;
	}

	void PrintMinimizationFunctional(std::ostream& out) const
	{
		out << "\t" << GetMinimizationFunctional() << "\n";
	}

private:
	std::string GetMinimizationFunctional() const
	{
		std::string f;
		for (auto& var : m_binVars)
		{
			if (var.front() == 't')
			{
				f += " + " + var;
			}
			else if (var.front() == 'p')
			{
				f += " - 2 " + var;
			}
			else if (var.front() == 'r')
			{
				f += " + 2 " + var;
			}
		}

		return f;
	}

private:
	std::set<std::string> m_binVars;
	std::set<std::string> m_genVars;
	std::map<std::string, int> m_mVars;
	std::map<std::string, int> m_byNodeVals;
};

#endif // _VARIABLES_STORAGE_H_
